/*
 * Copyright (C) 2013-2018 Nikos Mavrogiannopoulos
 * Copyright (C) 2015, 2016 Red Hat, Inc.
 *
 * This file is part of ocserv.
 *
 * ocserv is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * ocserv is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef OC_WORKER_LATENCY_H
#define OC_WORKER_LATENCY_H

ssize_t dtls_pull_latency(gnutls_transport_ptr_t ptr, void *data, size_t size);
void send_latency_stats_delta_to_main(worker_st *ws, time_t now);
void capture_latency_sample(struct worker_st *ws,
			    struct timespec *processing_start_time);

#endif
